// SPDX-FileCopyrightText: Matteo Settenvini <matteo.settenvini@montecristosoftware.eu>
// SPDX-License-Identifier: EUPL-1.2

use super::Cleaner;
use crate::decision::{Action, Decision};
use anyhow::{Context, Result};
use async_trait::async_trait;
use std::path::PathBuf;
use tokio::sync::{
    broadcast::{self, error::RecvError},
    mpsc,
};

pub struct ListCleaner {
    action_type: Action,
    list: PathBuf,
}

impl ListCleaner {
    pub fn new(action_type: Action, list: PathBuf) -> Self {
        Self { action_type, list }
    }
}

#[async_trait]
impl Cleaner for ListCleaner {
    async fn run(
        &mut self,
        mut files: broadcast::Receiver<PathBuf>,
        decisions: mpsc::Sender<Decision>,
    ) -> Result<()> {
        let current_dir = std::env::current_dir()?;

        // We use a gitignore builder because it has the syntax and
        // expressivity we need. Checking if a file "is ignored"
        // will tell us if it matches against the allow-/block-list
        let mut matcher_builder = ignore::gitignore::GitignoreBuilder::new(current_dir);
        if let Some(err) = matcher_builder.add(&self.list) {
            anyhow::bail!("list: error parsing {}: {}", self.list.display(), err);
        }

        let matcher = matcher_builder
            .build()
            .with_context(|| format!("list: cannot build matcher from {}", self.list.display()))?;

        let action_name = match self.action_type {
            Action::Keep => "allow",
            Action::Remove => "block",
        };

        loop {
            match files.recv().await {
                Ok(path) => {
                    if matcher.matched(&path, false).is_ignore() {
                        log::info!("{}list: marking {}", action_name, path.display());
                        decisions
                            .send(Decision {
                                action: self.action_type,
                                path: path,
                            })
                            .await?;
                    }
                }
                Err(RecvError::Closed) => break,
                Err(e) => Err(anyhow::anyhow!(e)).context("allowlist: recv error")?,
            }
        }

        Ok(())
    }
}
